import { AxiosInstance } from 'axios';
import { JsonApiResponse } from './response';
import { JsonApiClientConfig, JsonApiClientRequestOpts, JsonApiDefaultMeta, URL } from './types';
/**
 * Represents JsonApiClient that performs requests to backend
 */
export declare class JsonApiClient {
    private _baseUrl;
    private _axios;
    constructor(config?: JsonApiClientConfig);
    /**
     * Clones current JsonApiClient instance
     */
    private _clone;
    /**
     * Sets axios instance to the client instance.
     */
    get axios(): AxiosInstance;
    /**
     * Sets axios instance to the client instance.
     */
    useAxios(axiosInstance: AxiosInstance): JsonApiClient;
    /**
     *  Base URL will be prepended to `url` unless `url` is absolute.
     *  It can be convenient to set `baseURL` for an instance of axios to pass
     *  relative URLs to methods of that instance.
     *
     *  For more details look Axios Request config:
     *  {@link https://github.com/axios/axios#request-config}
     */
    get baseUrl(): URL;
    /**
     * Assigns new base URL to the current instance.
     */
    useBaseUrl(baseUrl: URL): JsonApiClient;
    /**
     * Creates new instance JsonApiClient instance with given base URL.
     */
    withBaseUrl(baseUrl: URL): JsonApiClient;
    /**
     * Performs a http request
     */
    request<T, U = JsonApiDefaultMeta>(opts: JsonApiClientRequestOpts): Promise<JsonApiResponse<T, U>>;
    /**
     * Makes a `GET` to a target `endpoint` with the provided `query` params.
     * Parses the response in JsonApi format.
     */
    get<T, U = JsonApiDefaultMeta>(endpoint: string, query?: Record<string, unknown>, isNeedRaw?: boolean): Promise<JsonApiResponse<T, U>>;
    /**
     * Makes a `POST` to a target `endpoint` with the provided `data` as body.
     * Parses the response in JsonApi format.
     */
    post<T, U = JsonApiDefaultMeta>(endpoint: string, data: unknown, isNeedRaw?: boolean): Promise<JsonApiResponse<T, U>>;
    /**
     * Makes a `PATCH` to a target `endpoint` with the provided `data` as body.
     * Signing can be enabled with `needSign` argument. Parses the response in
     * JsonApi format.
     */
    patch<T, U = JsonApiDefaultMeta>(endpoint: string, data?: unknown): Promise<JsonApiResponse<T, U>>;
    /**
     * Makes a `PUT` to a target `endpoint` with the provided `data` as body.
     * Parses the response in JsonApi format.
     */
    put<T, U = JsonApiDefaultMeta>(endpoint: string, data: unknown): Promise<JsonApiResponse<T, U>>;
    /**
     * Makes a `DELETE` to a target `endpoint` with the provided `data` as body.
     * Parses the response in JsonApi format.
     */
    delete<T, U = JsonApiDefaultMeta>(endpoint: string, data?: unknown): Promise<JsonApiResponse<T, U>>;
}
